from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StringType, StructField, StructType
from spetlrtools.testing.DataframeTestCase import DataframeTestCase

from spetlr import Configurator
from spetlr.functions import get_unique_tempview_name
from spetlr.utils import DataframeCreator
from tests.cluster.sql.DeliverySqlServer import DeliverySqlServer

from . import extras


class DeliverySqlServerTests(DataframeTestCase):
    tc = None
    sql_server = None
    table_name = get_unique_tempview_name()
    table_upsert_name = get_unique_tempview_name()
    view_name = get_unique_tempview_name()

    @classmethod
    def setUpClass(cls):
        cls.sql_server = DeliverySqlServer()
        cls.tc = Configurator()

        cls.tc.add_resource_path(extras)
        cls.tc.reset(debug=True)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.sql_server.drop_table_by_name(cls.table_name)
        cls.sql_server.drop_table_by_name(cls.table_upsert_name)
        t1 = cls.tc.table_name("SqlTestTable1")
        t2 = cls.tc.table_name("SqlTestTable2")
        v1 = cls.tc.table_name("SqlTestView")
        cls.sql_server.drop_table_by_name(t1)
        cls.sql_server.drop_table_by_name(t2)
        cls.sql_server.drop_view_by_name(v1)

        cls.tc.reset(debug=False)

    def test01_can_connect(self):
        self.sql_server.test_odbc_connection()
        self.assertTrue(True)

    def test02_can_create_dummy(self):
        self.create_test_table()
        self.assertTrue(True)

    def test03_can_read_dummy(self):
        self.sql_server.read_table_by_name(self.table_name)
        self.assertTrue(True)

    def test03_can_truncate_dummy(self):
        self.insert_single_row()
        df_with_data = self.sql_server.read_table_by_name(self.table_name)
        self.assertEqual(df_with_data.count(), 1)

        # Truncate
        self.sql_server.truncate_table_by_name(self.table_name)
        df_without_data = self.sql_server.read_table_by_name(self.table_name)
        self.assertEqual(df_without_data.count(), 0)

    def test04_can_load_sql_spark_dummy(self):
        sql_argument = f"""
                (select * from {self.table_name}) target
                                """
        self.insert_single_row()
        df_with_data = self.sql_server.load_sql(sql_argument)
        self.assertEqual(df_with_data.count(), 1)

    def test05_can_drop_table_dummy(self):
        self.sql_server.drop_table_by_name(self.table_name)

        sql_argument = f"""
        (SELECT * FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_NAME = '{self.table_name}') target
        """
        table_exists = self.sql_server.load_sql(sql_argument)
        self.assertEqual(table_exists.count(), 0)

    def test06_write_table_spark_dummy(self):
        # create test table again
        self.create_test_table()

        # write using spark
        df_export = self.create_data()
        self.sql_server.write_table_by_name(df_export, self.table_name)
        df_with_data = self.sql_server.read_table_by_name(self.table_name)
        self.assertEqual(df_with_data.count(), 1)

    def test08_get_table_name(self):
        test_name1 = self.sql_server.table_name("SqlTestTable1")
        self.assertIn("dbo.test1", test_name1)
        test_name2 = self.sql_server.table_name("SqlTestTable2")
        self.assertIn("dbo.test2", test_name2)

    def test09_execute_sql_file(self):
        file_name = "test1"
        path_name = extras

        # Create the table
        self.sql_server.execute_sql_file(resource_path=path_name, sql_file=file_name)
        self.assertTrue(True)

    def test10_read_w_id(self):
        # This might fail if the previous test didnt succeed
        self.sql_server.read_table("SqlTestTable1")
        self.sql_server.read_table("SqlTestTable2")
        self.assertTrue(True)

    def test11_write_w_id(self):
        df = self.create_data()
        self.sql_server.write_table(df, "SqlTestTable1")
        df_with_data = self.sql_server.read_table("SqlTestTable1")
        self.assertEqual(df_with_data.count(), 1)

    def test12_truncate_w_id(self):
        # Truncate
        self.sql_server.truncate_table("SqlTestTable1")
        df_without_data = self.sql_server.read_table("SqlTestTable1")
        self.assertEqual(df_without_data.count(), 0)

    def test13_drop_w_id(self):
        self.sql_server.drop_table("SqlTestTable1")

        table1_name = self.tc.table_name("SqlTestTable1")
        sql_argument = f"""
                (SELECT * FROM INFORMATION_SCHEMA.TABLES
                WHERE TABLE_NAME = '{table1_name}') target
                """
        table_exists = self.sql_server.load_sql(sql_argument)
        self.assertEqual(table_exists.count(), 0)

    def test14_can_drop_view_w_id(self):
        # Create view
        view_name = self.tc.table_name("SqlTestView")
        self.assertIn("viewtest1", view_name)
        table_from = self.tc.table_name("SqlTestTable2")
        self.create_test_view(view_name, table_from)
        self.assertTrue(True)

        # Drop view by id
        self.sql_server.drop_view("SqlTestView")

        sql_argument = f"""
        (select
                    *
                    from
                    INFORMATION_SCHEMA.VIEWS
                    where
                    table_name = '{view_name}') target
        """
        table_exists = self.sql_server.load_sql(sql_argument)
        self.assertEqual(table_exists.count(), 0)

    def test15_can_drop_view_by_name(self):
        table_from = self.tc.table_name("SqlTestTable2")
        self.create_test_view(self.view_name, table_from)
        self.sql_server.drop_view_by_name(self.view_name)
        sql_argument = f"""
                (select
                            *
                            from
                            INFORMATION_SCHEMA.VIEWS
                            where
                            table_name = '{self.view_name}') target
                """
        table_exists = self.sql_server.load_sql(sql_argument)
        self.assertEqual(table_exists.count(), 0)

    def test16_upsert_to_table_none_input(self):
        val_return = self.sql_server.upsert_to_table_by_name(
            df_source=None,
            table_name=self.table_upsert_name,
            join_cols=["testid"],
            filter_join_cols=False,
        )

        self.assertEqual(val_return, None)

    def test17_upsert_to_table(self):
        self.create_upsert_test_table()

        upsertTableSchema = StructType(
            [
                StructField("testid", IntegerType(), True),
                StructField("testdata", StringType(), True),
            ]
        )

        # Write pre upsert test data
        df_preTest = DataframeCreator.make_partial(
            schema=upsertTableSchema,
            columns=["testid", "testdata"],
            data=[(1, "testdata1"), (2, "testdata2")],
        )
        self.sql_server.write_table_by_name(
            df_source=df_preTest, table_name=self.table_upsert_name, append=False
        )

        # Write upsert test data
        df_upsertTest = DataframeCreator.make_partial(
            schema=upsertTableSchema,
            columns=["testid", "testdata"],
            data=[(2, "newtestdata2"), (3, "testdata3")],
        )
        self.sql_server.upsert_to_table_by_name(
            df_source=df_upsertTest,
            table_name=self.table_upsert_name,
            join_cols=["testid"],
            filter_join_cols=False,
            overwrite_if_target_is_empty=False,
        )

        # Validate correct upserted data
        df_afterUpsert = self.sql_server.read_table_by_name(self.table_upsert_name)
        expectedData = [(1, "testdata1"), (2, "newtestdata2"), (3, "testdata3")]

        self.assertDataframeMatches(
            df=df_afterUpsert,
            expected_data=expectedData,
        )

    def test18_upsert_to_table_with_join_cols_filter(self):
        upsertTableSchema = StructType(
            [
                StructField("testid", IntegerType(), True),
                StructField("testdata", StringType(), True),
            ]
        )

        # Write pre upsert test data
        df_preTest = DataframeCreator.make_partial(
            schema=upsertTableSchema,
            columns=["testid", "testdata"],
            data=[(1, "testdata1"), (2, "testdata2")],
        )
        self.sql_server.write_table_by_name(
            df_source=df_preTest, table_name=self.table_upsert_name, append=False
        )

        # Write upsert test data with None keys
        df_upsertTest = DataframeCreator.make_partial(
            schema=upsertTableSchema,
            columns=["testid", "testdata"],
            data=[(2, "newtestdata2"), (3, "testdata3"), (None, "testdata4")],
        )
        self.sql_server.upsert_to_table_by_name(
            df_source=df_upsertTest,
            table_name=self.table_upsert_name,
            join_cols=["testid"],
            filter_join_cols=True,
        )

        # Validate correct upserted data
        df_afterUpsert = self.sql_server.read_table_by_name(self.table_upsert_name)
        expectedData = [(1, "testdata1"), (2, "newtestdata2"), (3, "testdata3")]

        self.assertDataframeMatches(
            df=df_afterUpsert,
            expected_data=expectedData,
        )

    def test19_upsert_to_table_overwrite_empty_target(self):
        upsertTableSchema = StructType(
            [
                StructField("testid", IntegerType(), True),
                StructField("testdata", StringType(), True),
            ]
        )

        # Write empty pre upsert test data
        df_preTest = DataframeCreator.make_partial(
            schema=upsertTableSchema,
            columns=["testid", "testdata"],
            data=[],
        )
        self.sql_server.write_table_by_name(
            df_source=df_preTest, table_name=self.table_upsert_name, append=False
        )

        # Write upsert test data
        df_upsertTest = DataframeCreator.make_partial(
            schema=upsertTableSchema,
            columns=["testid", "testdata"],
            data=[(2, "newtestdata2"), (3, "testdata3")],
        )
        self.sql_server.upsert_to_table_by_name(
            df_source=df_upsertTest,
            table_name=self.table_upsert_name,
            join_cols=["testid"],
            overwrite_if_target_is_empty=True,
        )

        # Validate correct upserted data
        df_afterUpsert = self.sql_server.read_table_by_name(self.table_upsert_name)
        expectedData = [(2, "newtestdata2"), (3, "testdata3")]

        self.assertDataframeMatches(
            df=df_afterUpsert,
            expected_data=expectedData,
        )

    def test20_big_data_set(self):
        # Ensure table creation
        self.sql_server.drop_table_by_name(self.table_name)
        self.create_test_table()

        # Create test data
        df = self.create_data()

        # Write as big data set
        self.sql_server.write_table_by_name(df, self.table_name, big_data_set=True)
        df_with_data = self.sql_server.read_table_by_name(self.table_name)
        self.assertEqual(df_with_data.count(), 1)

    def create_test_table(self):
        sql_argument = f"""
            IF OBJECT_ID('{self.table_name}', 'U') IS NULL
            BEGIN
            CREATE TABLE {self.table_name}
            (
                testcolumn INT NULL
            )
            END
        """
        self.sql_server.execute_sql(sql_argument)

    def insert_single_row(self):
        insert_data = 123

        sql_argument = f"""
            INSERT INTO {self.table_name} values ({insert_data})
        """
        self.sql_server.execute_sql(sql_argument)

    def create_upsert_test_table(self):
        sql_argument = f"""
            IF OBJECT_ID('{self.table_upsert_name}', 'U') IS NULL
            BEGIN
            CREATE TABLE {self.table_upsert_name}
            (
                testid INT NULL,
                testdata nvarchar(max) NULL
            )
            END
        """
        self.sql_server.execute_sql(sql_argument)

    def create_data(self) -> DataFrame:
        schema = StructType(
            [
                StructField("testcolumn", IntegerType(), True),
            ]
        )
        cols = ["testcolumn"]
        df_new = DataframeCreator.make_partial(
            schema=schema,
            columns=cols,
            data=[(456,)],
        )

        return df_new.orderBy("testcolumn")

    def create_test_view(self, view_name, select_from_table):
        sql_argument = f"""
            CREATE OR ALTER VIEW {view_name} as
            (
            select * from {select_from_table}
            )
        """
        self.sql_server.execute_sql(sql_argument)



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.exceptions import (
    ColumnDoesNotExistException,
    MoreThanTwoDataFramesException,
)
from spetlr.spark import Spark
from spetlr.transformers import JoinDataframesTransformerNC


class TestJoinDataframesTransformerNC(DataframeTestCase):
    def test_data_join_inner(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        df_transformed = JoinDataframesTransformerNC(
            first_dataframe_join_key="Id",
            second_dataframe_join_key="_Id",
            join_type="inner",
            dataset_input_keys=["input1_df", "input2_df"],
        ).process_many(dataset)

        expected_data = [
            (
                "Id1",
                "Name1",
                "Id1",
                "Country1",
            ),
            (
                "Id2",
                "Name2",
                "Id2",
                "Country2",
            ),
        ]

        self.assertDataframeMatches(
            df=df_transformed,
            expected_data=expected_data,
        )

    def test_data_join_outer(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        df_transformed = JoinDataframesTransformerNC(
            first_dataframe_join_key="Id",
            second_dataframe_join_key="_Id",
            join_type="outer",
            dataset_input_keys=["input1_df", "input2_df"],
        ).process_many(dataset)

        transformed_data = [tuple(row) for row in df_transformed.collect()]

        expected_data = [
            (
                "Id1",
                "Name1",
                "Id1",
                "Country1",
            ),
            (
                "Id2",
                "Name2",
                "Id2",
                "Country2",
            ),
            (None, None, "Id3", "Country3"),
        ]

        self.assertEqual(
            first=transformed_data,
            second=expected_data,
        )

    def test_data_join_left_outer(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
            ("Id4", "Name4"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        df_transformed = JoinDataframesTransformerNC(
            first_dataframe_join_key="Id",
            second_dataframe_join_key="_Id",
            join_type="left_outer",
            dataset_input_keys=["input1_df", "input2_df"],
        ).process_many(dataset)

        expected_data = [
            (
                "Id1",
                "Name1",
                "Id1",
                "Country1",
            ),
            (
                "Id2",
                "Name2",
                "Id2",
                "Country2",
            ),
            ("Id4", "Name4", None, None),
        ]

        self.assertDataframeMatches(
            df=df_transformed,
            expected_data=expected_data,
        )

    def test_data_join_right_outer(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        df_transformed = JoinDataframesTransformerNC(
            first_dataframe_join_key="Id",
            second_dataframe_join_key="_Id",
            join_type="right_outer",
            dataset_input_keys=["input1_df", "input2_df"],
        ).process_many(dataset)

        transformed_data = [tuple(row) for row in df_transformed.collect()]

        expected_data = [
            (
                "Id1",
                "Name1",
                "Id1",
                "Country1",
            ),
            (
                "Id2",
                "Name2",
                "Id2",
                "Country2",
            ),
            (None, None, "Id3", "Country3"),
        ]

        self.assertEqual(
            first=transformed_data,
            second=expected_data,
        )

    def test_data_join_semi(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        df_transformed = JoinDataframesTransformerNC(
            first_dataframe_join_key="Id",
            second_dataframe_join_key="_Id",
            join_type="semi",
            dataset_input_keys=["input1_df", "input2_df"],
        ).process_many(dataset)

        expected_data = [
            (
                "Id1",
                "Name1",
            ),
            (
                "Id2",
                "Name2",
            ),
        ]

        self.assertDataframeMatches(
            df=df_transformed,
            expected_data=expected_data,
        )

    def test_data_join_anti(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
            ("Id4", "Name4"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        df_transformed = JoinDataframesTransformerNC(
            first_dataframe_join_key="Id",
            second_dataframe_join_key="_Id",
            join_type="anti",
            dataset_input_keys=["input1_df", "input2_df"],
        ).process_many(dataset)

        expected_data = [("Id4", "Name4")]

        self.assertDataframeMatches(
            df=df_transformed,
            expected_data=expected_data,
        )

    def test_data_join_MoreThanTwoDataFramesException(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
            ("Id4", "Name4"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        input3_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Age", T.StringType(), True),
            ]
        )

        input3_data = [
            ("Id1", "Age1"),
            ("Id2", "Age2"),
            ("Id3", "Age3"),
        ]

        dataset["input3_df"] = Spark.get().createDataFrame(
            schema=input3_schema, data=input3_data
        )

        with self.assertRaises(MoreThanTwoDataFramesException):
            JoinDataframesTransformerNC(
                first_dataframe_join_key="Id",
                second_dataframe_join_key="_Id",
                join_type="anti",
                dataset_input_keys=["input1_df", "input2_df", "input3_df"],
            ).process_many(dataset)

    def test_data_join_ColumnDoesNotExistException(self):
        dataset = {}

        input1_schema = T.StructType(
            [
                T.StructField("Id", T.StringType(), True),
                T.StructField("Name", T.StringType(), True),
            ]
        )

        input1_data = [
            ("Id1", "Name1"),
            ("Id2", "Name2"),
            ("Id4", "Name4"),
        ]

        dataset["input1_df"] = Spark.get().createDataFrame(
            schema=input1_schema, data=input1_data
        )

        input2_schema = T.StructType(
            [
                T.StructField("_Id", T.StringType(), True),
                T.StructField("Country", T.StringType(), True),
            ]
        )

        input2_data = [
            ("Id1", "Country1"),
            ("Id2", "Country2"),
            ("Id3", "Country3"),
        ]

        dataset["input2_df"] = Spark.get().createDataFrame(
            schema=input2_schema, data=input2_data
        )

        with self.assertRaises(ColumnDoesNotExistException):
            JoinDataframesTransformerNC(
                first_dataframe_join_key="Id_col",
                second_dataframe_join_key="_Id",
                join_type="anti",
                dataset_input_keys=["input1_df", "input2_df"],
            ).process_many(dataset)

        with self.assertRaises(ColumnDoesNotExistException):
            JoinDataframesTransformerNC(
                first_dataframe_join_key="Id",
                second_dataframe_join_key="_Id_col",
                join_type="anti",
                dataset_input_keys=["input1_df", "input2_df"],
            ).process_many(dataset)



import unittest

from pyspark.sql.utils import AnalysisException

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors import SimpleExtractor
from spetlr.etl.extractors.schema_extractor import SchemaExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.spark import Spark


class DeltaTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()

    def test_01_configure(self):
        tc = Configurator()
        tc.register(
            "MyDb", {"name": "TestDb{ID}", "path": "/mnt/spetlr/silver/testdb{ID}"}
        )

        tc.register(
            "MyTbl",
            {
                "name": "TestDb{ID}.TestTbl",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl",
            },
        )

        tc.register(
            "MyTbl2",
            {
                "name": "TestDb{ID}.TestTbl2",
            },
        )

        tc.register(
            "MyTbl3",
            {
                "path": "/mnt/spetlr/silver/testdb/testtbl3",
            },
        )

        # test instantiation without error
        DbHandle.from_tc("MyDb")
        DeltaHandle.from_tc("MyTbl")
        DeltaHandle.from_tc("MyTbl2")

    def test_02_write(self):
        dh = DeltaHandle.from_tc("MyTbl")

        df = Spark.get().createDataFrame([(1, "a"), (2, "b")], "id int, name string")

        dh.overwrite(df, mergeSchema=True)
        dh.append(df, mergeSchema=False)  # schema matches

        df = Spark.get().createDataFrame(
            [(1, "a", "yes"), (2, "b", "no")],
            """
            id int,
            name string,
            response string
            """,
        )

        dh.append(df, mergeSchema=True)

    def test_03_create(self):
        db = DbHandle.from_tc("MyDb")
        db.create()

        dh = DeltaHandle.from_tc("MyTbl")
        dh.create_hive_table()

        # test hive access:
        df = Spark.get().table("TestDb.TestTbl")
        self.assertTrue(6, df.count())

    def test_04_read(self):
        df = DeltaHandle.from_tc("MyTbl").read()
        self.assertEqual(6, df.count())

    def test_041_schema(self):
        """Schema Extractor returns zero lines, but preserves the schema,
        giving more readable orchestrators"""
        df1 = SimpleExtractor(DeltaHandle.from_tc("MyTbl"), "MyTbl").read()
        self.assertEqual(6, df1.count())

        df2 = SchemaExtractor(DeltaHandle.from_tc("MyTbl"), "MyTbl").read()
        self.assertEqual(0, df2.count())

        self.assertEqual(df1.schema, df2.schema)

    def test_05_truncate(self):
        dh = DeltaHandle.from_tc("MyTbl")
        dh.truncate()
        df = dh.read()
        self.assertEqual(0, df.count())

    def test_06_etl(self):
        o = Orchestrator()
        o.extract_from(
            SimpleExtractor(DeltaHandle.from_tc("MyTbl"), dataset_key="MyTbl")
        )
        o.load_into(SimpleLoader(DeltaHandle.from_tc("MyTbl"), mode="overwrite"))
        o.execute()

    def test_07_write_path_only(self):
        # check that we can write to the table with no path
        df = DeltaHandle.from_tc("MyTbl").read()

        dh3 = DeltaHandle.from_tc("MyTbl3")

        dh3.append(df, mergeSchema=True)

        df = dh3.read()
        df.show()

    def test_08_delete(self):
        dh = DeltaHandle.from_tc("MyTbl")
        dh.drop_and_delete()

        with self.assertRaises(AnalysisException):
            dh.read()

    def test_09_partitioning(self):
        dh = DeltaHandle.from_tc("MyTbl")
        Spark.get().sql(
            f"""
            CREATE TABLE {dh.get_tablename()}
            (
            colA string,
            colB int,
            payload string
            )
            PARTITIONED BY (colB,colA)
        """
        )

        self.assertEqual(dh.get_partitioning(), ["colB", "colA"])

        dh2 = DeltaHandle.from_tc("MyTbl2")
        Spark.get().sql(
            f"""
            CREATE TABLE {dh2.get_tablename()}
            (
            colA string,
            colB int,
            payload string
            )
        """
        )

        self.assertEqual(dh2.get_partitioning(), [])



class SpetlrException(Exception):
    pass


class SpetlrKeyError(KeyError):
    pass


class NoTableException(SpetlrException):
    value = "No table found!"
    pass


class UnkownPathException(SpetlrException):
    value = "Something went wrong during reading of path!"
    pass


class ColumnDoesNotExistException(SpetlrException):
    pass


class MoreThanTwoDataFramesException(SpetlrException):
    pass


class EhJsonToDeltaException(SpetlrException):
    pass


class NoSuchSchemaException(SpetlrException):
    pass


class FalseSchemaDefinitionException(SpetlrException):
    pass


class UnregisteredSchemaDefinitionException(SpetlrException):
    pass


class NoRunId(SpetlrException):
    pass


class NoDbUtils(SpetlrException):
    pass


class NoSuchValueException(SpetlrKeyError):
    pass



import unittest
import warnings

from pyspark.sql import DataFrame

from spetlr.etl import Extractor, Orchestrator


class OrchestratorEtlTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        warnings.filterwarnings("error")

    def test_no_special_case(self):
        class MyEx(Extractor):
            def read(self) -> DataFrame:
                return "hi"

        class MyO(Orchestrator):
            def __init__(self):
                super().__init__()
                self.extract_from(MyEx())

        warnings.filterwarnings("error")
        # no warning
        _ = MyO().execute()

    def test_misuse_warning(self):
        class MyEx(Extractor):
            def read(self) -> DataFrame:
                return "hi"

        class MyO(Orchestrator):
            def __init__(self):
                super().__init__()
                self.extract_from(MyEx())

        with self.assertRaises(UserWarning):
            warnings.filterwarnings("error")
            _ = MyO().execute({"input": "foobar"})

    def test_misuse_warning_suppressed(self):
        class MyEx(Extractor):
            def read(self) -> DataFrame:
                return "hi"

        class MyO(Orchestrator):
            def __init__(self):
                super().__init__(suppress_composition_warning=True)
                self.extract_from(MyEx())

        warnings.filterwarnings("error")
        # no warning
        _ = MyO().execute()

    def test_correct_use_no_warning(self):
        class MyEx(Extractor):
            def read(self) -> DataFrame:
                if len(self.previous_extractions):
                    return self.previous_extractions.popitem()[1]
                return "hi"

        class MyO(Orchestrator):
            def __init__(self):
                super().__init__()
                self.extract_from(MyEx())

        warnings.filterwarnings("error")
        # no warning
        _ = MyO().execute()


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""SQL formatter"""

from spetlr.configurator.sql.sqlparse import filters
from spetlr.configurator.sql.sqlparse.exceptions import SQLParseError


def validate_options(options):
    """Validates options."""
    kwcase = options.get("keyword_case")
    if kwcase not in [None, "upper", "lower", "capitalize"]:
        raise SQLParseError("Invalid value for keyword_case: " "{!r}".format(kwcase))

    idcase = options.get("identifier_case")
    if idcase not in [None, "upper", "lower", "capitalize"]:
        raise SQLParseError("Invalid value for identifier_case: " "{!r}".format(idcase))

    ofrmt = options.get("output_format")
    if ofrmt not in [None, "sql", "python", "php"]:
        raise SQLParseError("Unknown output format: " "{!r}".format(ofrmt))

    strip_comments = options.get("strip_comments", False)
    if strip_comments not in [True, False]:
        raise SQLParseError(
            "Invalid value for strip_comments: " "{!r}".format(strip_comments)
        )

    space_around_operators = options.get("use_space_around_operators", False)
    if space_around_operators not in [True, False]:
        raise SQLParseError(
            "Invalid value for use_space_around_operators: "
            "{!r}".format(space_around_operators)
        )

    strip_ws = options.get("strip_whitespace", False)
    if strip_ws not in [True, False]:
        raise SQLParseError(
            "Invalid value for strip_whitespace: " "{!r}".format(strip_ws)
        )

    truncate_strings = options.get("truncate_strings")
    if truncate_strings is not None:
        try:
            truncate_strings = int(truncate_strings)
        except (ValueError, TypeError):
            raise SQLParseError(
                "Invalid value for truncate_strings: " "{!r}".format(truncate_strings)
            )
        if truncate_strings <= 1:
            raise SQLParseError(
                "Invalid value for truncate_strings: " "{!r}".format(truncate_strings)
            )
        options["truncate_strings"] = truncate_strings
        options["truncate_char"] = options.get("truncate_char", "[...]")

    indent_columns = options.get("indent_columns", False)
    if indent_columns not in [True, False]:
        raise SQLParseError(
            "Invalid value for indent_columns: " "{!r}".format(indent_columns)
        )
    elif indent_columns:
        options["reindent"] = True  # enforce reindent
    options["indent_columns"] = indent_columns

    reindent = options.get("reindent", False)
    if reindent not in [True, False]:
        raise SQLParseError("Invalid value for reindent: " "{!r}".format(reindent))
    elif reindent:
        options["strip_whitespace"] = True

    reindent_aligned = options.get("reindent_aligned", False)
    if reindent_aligned not in [True, False]:
        raise SQLParseError(
            "Invalid value for reindent_aligned: " "{!r}".format(reindent)
        )
    elif reindent_aligned:
        options["strip_whitespace"] = True

    indent_after_first = options.get("indent_after_first", False)
    if indent_after_first not in [True, False]:
        raise SQLParseError(
            "Invalid value for indent_after_first: " "{!r}".format(indent_after_first)
        )
    options["indent_after_first"] = indent_after_first

    indent_tabs = options.get("indent_tabs", False)
    if indent_tabs not in [True, False]:
        raise SQLParseError(
            "Invalid value for indent_tabs: " "{!r}".format(indent_tabs)
        )
    elif indent_tabs:
        options["indent_char"] = "\t"
    else:
        options["indent_char"] = " "

    indent_width = options.get("indent_width", 2)
    try:
        indent_width = int(indent_width)
    except (TypeError, ValueError):
        raise SQLParseError("indent_width requires an integer")
    if indent_width < 1:
        raise SQLParseError("indent_width requires a positive integer")
    options["indent_width"] = indent_width

    wrap_after = options.get("wrap_after", 0)
    try:
        wrap_after = int(wrap_after)
    except (TypeError, ValueError):
        raise SQLParseError("wrap_after requires an integer")
    if wrap_after < 0:
        raise SQLParseError("wrap_after requires a positive integer")
    options["wrap_after"] = wrap_after

    comma_first = options.get("comma_first", False)
    if comma_first not in [True, False]:
        raise SQLParseError("comma_first requires a boolean value")
    options["comma_first"] = comma_first

    right_margin = options.get("right_margin")
    if right_margin is not None:
        try:
            right_margin = int(right_margin)
        except (TypeError, ValueError):
            raise SQLParseError("right_margin requires an integer")
        if right_margin < 10:
            raise SQLParseError("right_margin requires an integer > 10")
    options["right_margin"] = right_margin

    return options


def build_filter_stack(stack, options):
    """Setup and return a filter stack.

    Args:
      stack: :class:`~sqlparse.filters.FilterStack` instance
      options: Dictionary with options validated by validate_options.
    """
    # Token filter
    if options.get("keyword_case"):
        stack.preprocess.append(filters.KeywordCaseFilter(options["keyword_case"]))

    if options.get("identifier_case"):
        stack.preprocess.append(
            filters.IdentifierCaseFilter(options["identifier_case"])
        )

    if options.get("truncate_strings"):
        stack.preprocess.append(
            filters.TruncateStringFilter(
                width=options["truncate_strings"], char=options["truncate_char"]
            )
        )

    if options.get("use_space_around_operators", False):
        stack.enable_grouping()
        stack.stmtprocess.append(filters.SpacesAroundOperatorsFilter())

    # After grouping
    if options.get("strip_comments"):
        stack.enable_grouping()
        stack.stmtprocess.append(filters.StripCommentsFilter())

    if options.get("strip_whitespace") or options.get("reindent"):
        stack.enable_grouping()
        stack.stmtprocess.append(filters.StripWhitespaceFilter())

    if options.get("reindent"):
        stack.enable_grouping()
        stack.stmtprocess.append(
            filters.ReindentFilter(
                char=options["indent_char"],
                width=options["indent_width"],
                indent_after_first=options["indent_after_first"],
                indent_columns=options["indent_columns"],
                wrap_after=options["wrap_after"],
                comma_first=options["comma_first"],
            )
        )

    if options.get("reindent_aligned", False):
        stack.enable_grouping()
        stack.stmtprocess.append(
            filters.AlignedIndentFilter(char=options["indent_char"])
        )

    if options.get("right_margin"):
        stack.enable_grouping()
        stack.stmtprocess.append(
            filters.RightMarginFilter(width=options["right_margin"])
        )

    # Serializer
    if options.get("output_format"):
        frmt = options["output_format"]
        if frmt.lower() == "php":
            fltr = filters.OutputPHPFilter()
        elif frmt.lower() == "python":
            fltr = filters.OutputPythonFilter()
        else:
            fltr = None
        if fltr is not None:
            stack.postprocess.append(fltr)

    return stack



import json
from functools import lru_cache
from typing import Dict

from requests import HTTPError

from spetlr.db_auto import getDbApi
from spetlr.exceptions import NoDbUtils, NoRunId
from spetlr.functions import init_dbutils


class JobReflection:
    @classmethod
    @lru_cache
    def get_job_details(cls) -> Dict:
        try:
            # this strange construct reaches beyond the pyspark api layer
            # and into undocumented structures in order to retrieve the details
            # of the current run.
            # if you know a better way, please add it here.
            return json.loads(
                init_dbutils()
                .notebook.entry_point.getDbutils()
                .notebook()
                .getContext()
                .toJson()
            )
        except (AttributeError, NoDbUtils):
            raise NoRunId()

    @classmethod
    @lru_cache
    def get_job_api_details(cls) -> Dict:
        run_id = cls.get_current_run_id()
        db_api = getDbApi()

        try:
            return db_api.jobs.get_run(run_id=run_id)
        except HTTPError:
            raise NoRunId("DbApi returned http error.")

    @classmethod
    def get_job_results_url(cls) -> str:
        return cls.get_job_api_details()["run_page_url"]

    @classmethod
    def get_job_name(cls) -> str:
        return cls.get_job_api_details()["run_name"]

    @classmethod
    @lru_cache
    def get_current_run_id(cls) -> int:
        details = cls.get_job_details()
        try:
            # the first part is set in a task orchestration job
            # the second one is set in a plain old run
            run = details.get("currentRunId", None) or details["rootRunId"]
            return int(run["id"])  # convert to int for safety
        except (TypeError, KeyError):
            raise NoRunId()



import unittest

import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.spark import Spark
from spetlr.transformers import SelectAndCastColumnsTransformerNC


class TestSelectAndCastColumnsTransformerNC(DataframeTestCase):
    def test_select_transformer(self):
        inputSchema = T.StructType(
            [
                T.StructField("id", T.IntegerType(), True),
                T.StructField("value", T.StringType(), True),
            ]
        )

        inputData = [
            (
                42,
                "data",
            )
        ]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        transformer_schema = T.StructType(
            [
                T.StructField("id", T.IntegerType(), True),
                T.StructField("value2", T.StringType(), True),
            ]
        )

        transformed_df = SelectAndCastColumnsTransformerNC(
            schema=transformer_schema
        ).process(input_df)

        expectedData = [
            (
                42,
                None,
            )
        ]

        self.assertEqualSchema(transformed_df.schema, transformer_schema)
        self.assertDataframeMatches(transformed_df, None, expectedData)

    def test_cast_transformer(self):
        inputSchema = T.StructType(
            [
                T.StructField("id", T.IntegerType(), True),
                T.StructField("value", T.StringType(), True),
            ]
        )

        inputData = [
            (
                42,
                "data",
            )
        ]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        transformer_schema = T.StructType(
            [
                T.StructField("id", T.StringType(), True),
                T.StructField("value", T.StringType(), True),
            ]
        )

        transformed_df = SelectAndCastColumnsTransformerNC(
            schema=transformer_schema
        ).process(input_df)

        expectedData = [
            (
                "42",
                "data",
            )
        ]

        self.assertEqualSchema(transformed_df.schema, transformer_schema)
        self.assertDataframeMatches(transformed_df, None, expectedData)


if __name__ == "__main__":
    unittest.main()



import argparse
import os.path

from spetlr.exceptions.cli_exceptions import SpetlrCliCheckFailed


def setup_parser(parser: argparse.ArgumentParser):
    parser.set_defaults(func=generate_keys_file)
    parser.add_argument("-o", "--output-file", type=str, default="")
    parser.add_argument(
        "-c",
        "--check-only",
        dest="check",
        action="store_true",
        help="Only check, don't update the output file",
    )
    parser.set_defaults(check=False)


def generate_keys_file(self, options):
    """Generate a keys file from configured keys.
    The arguments are as follows:
        - self refers to the configurator object
        - options is a SimpleNamespace, expected to contain these attributes
            - output_file - str - the name of the output file, if none given,
              the generate file contents are printed to the console
            - check True/False, don't generate the file, verify an existing one.
    """
    new_conts = "\n".join(
        ["# AUTO GENERATED FILE", "# contains all spetlr.Configurator keys", ""]
        + [f"{key} = {repr(key)}" for key in sorted(self.all_keys())]
    )

    # if black is installed, use it to format the contents
    try:
        import black

        new_conts = black.format_file_contents(
            new_conts, fast=False, mode=black.FileMode()
        )
    except ModuleNotFoundError:
        pass

    if not options.output_file:
        # without output file, output to console
        print(new_conts, end="")
        return

    if not options.check:
        with open(options.output_file, "w") as f:
            f.write(new_conts)
        return

    if not os.path.exists(options.output_file):
        raise SpetlrCliCheckFailed(f"Output file {options.output_file} does not exist.")

    old_conts = open(options.output_file).read()
    if new_conts != old_conts:
        raise SpetlrCliCheckFailed(
            f"Output file {options.output_file} does not have correct contents."
        )

    # all checks passed. Return without error.
    return



from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from spetlr.eh import EventHubStream
from spetlr.etl import Loader


class EventHubJsonPublisher(Loader):
    """Save the rows of the incoming dataframe to the eventhub stream,
    formatted as json documents."""

    def __init__(self, eh: EventHubStream):
        super().__init__()
        self.eh = eh

    def save(self, df: DataFrame) -> None:
        self.eh.save_data(
            df.select(
                f.encode(
                    f.to_json(f.struct("*")),
                    "utf-8",
                ).alias("body")
            )
        )



import unittest
from textwrap import dedent

from pyspark.sql import types as t

from spetlr.schema_manager.spark_schema import get_schema


class TestGetSchema(unittest.TestCase):
    def test_01_schema1(self):
        sql = dedent(
            r"""
            a int NOT
            NULL,
            b int COMMENT "really? is that it?",
            c string,
            cplx struct< -- irrelevant comment
                someId:string,
                QrCode:string,
                details:struct/* why?! */<id:string>,
                blabla : array< int >
                >,
            d timestamp,
            m map<int,string>,
            p decimal(10,3),
            final string,
            gen DATE GENERATED ALWAYS AS (CAST(d AS DATE))
            """
        )
        struct = get_schema(sql)
        self.assertEqual(
            t.StructType(
                [
                    t.StructField("a", t.IntegerType(), False),
                    t.StructField(
                        "b",
                        t.IntegerType(),
                        True,
                        metadata={"comment": "really? is that it?"},
                    ),
                    t.StructField("c", t.StringType(), True),
                    t.StructField(
                        "cplx",
                        t.StructType(
                            [
                                t.StructField("someId", t.StringType(), True),
                                t.StructField("QrCode", t.StringType(), True),
                                t.StructField(
                                    "details",
                                    t.StructType(
                                        [t.StructField("id", t.StringType(), True)]
                                    ),
                                    True,
                                ),
                                t.StructField(
                                    "blabla", t.ArrayType(t.IntegerType(), True), True
                                ),
                            ]
                        ),
                        True,
                    ),
                    t.StructField("d", t.TimestampType(), True),
                    t.StructField(
                        "m", t.MapType(t.IntegerType(), t.StringType(), True), True
                    ),
                    t.StructField("p", t.DecimalType(10, 3), True),
                    t.StructField("final", t.StringType(), True),
                    t.StructField("gen", t.DateType(), True),
                ]
            ).json(),
            struct.json(),
        )


if __name__ == "__main__":
    unittest.main()



from .drop_oldest_duplicate_transformer import (  # noqa: F401
    DropOldestDuplicatesTransformer,
)
from .dropColumnsTransformer_nc import DropColumnsTransformerNC  # noqa: F401
from .join_dataframes_transformer import JoinDataframesTransformerNC  # noqa: F401
from .select_and_cast_columns_transformer_nc import (  # noqa: F401
    SelectAndCastColumnsTransformerNC,
)
from .selectColumnsTransformer_nc import SelectColumnsTransformerNC  # noqa: F401
from .simple_sql_transformer import SimpleSqlServerTransformer  # noqa: F401
from .timezone_transformer_nc import TimeZoneTransformerNC  # noqa: F401
from .union_transformer import UnionTransformer  # noqa: F401
from .union_transformer_nc import UnionTransformerNC  # noqa: F401



from typing import List, Protocol, Union

from pyspark.sql import DataFrame

from spetlr.etl import Loader


class Overwritable(Protocol):
    def overwrite(self, df: DataFrame) -> None:
        pass


class Appendable(Protocol):
    def append(self, df: DataFrame) -> None:
        pass


class SimpleLoader(Loader):
    def __init__(
        self,
        handle: Union[Overwritable, Appendable],
        *,
        mode: str = "overwrite",
        dataset_input_keys: Union[str, List[str]] = None,
    ):
        super().__init__(dataset_input_keys=dataset_input_keys)
        self.mode = mode
        self.handle = handle

    def save(self, df: DataFrame) -> None:
        if self.mode == "overwrite":
            self.handle.overwrite(df)
        else:
            self.handle.append(df)



import unittest

from pyspark.sql import DataFrame

from spetlr.etl import Orchestrator
from spetlr.etl.extractors import SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.spark import Spark
from spetlr.transformers import UnionTransformer


class MergeDfIntoTargetTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    def test_01_union(self):
        schema = "i integer, s string"

        class Reader:
            df: DataFrame

            def read(self):
                return self.df

        th1 = Reader()
        th1.df = Spark.get().createDataFrame([(1, "a")], schema)

        th2 = Reader()
        th2.df = Spark.get().createDataFrame([(1, "a")], schema)

        class Writer:
            df: DataFrame

            def overwrite(self, df: DataFrame):
                self.df = df

        th3 = Writer()

        o = Orchestrator()
        o.extract_from(SimpleExtractor(th1, "th1"))
        o.extract_from(SimpleExtractor(th2, "th2"))
        o.transform_with(UnionTransformer())
        o.load_into(SimpleLoader(th3))
        o.execute()

        self.assertEqual(th1.df.schema, th3.df.schema)



from typing import List


def GetMergeStatement(
    *,
    merge_statement_type: str,
    target_table_name: str,
    source_table_name: str,
    join_cols: List[str],
    insert_cols: List[str] = None,
    update_cols: List[str] = None,
    special_update_set: str = None,
) -> str:
    assert merge_statement_type in {"delta", "sql"}

    merge_sql_statement = (
        f"MERGE {'INTO ' if merge_statement_type == 'delta' else ''}"
        f"{target_table_name} AS target "
        f"USING {source_table_name} AS source "
        f"ON {' AND '.join(f'(source.{col} = target.{col})' for col in join_cols)} "
    )

    if update_cols and len(update_cols) > 0:
        merge_sql_statement += (
            "WHEN MATCHED THEN UPDATE "
            f"SET {', '.join(f'target.{col} = source.{col}' for col in update_cols)}"
        )

        if special_update_set:
            merge_sql_statement += f"{special_update_set} "
        else:
            merge_sql_statement += " "

    if insert_cols and len(insert_cols) > 0:
        merge_sql_statement += (
            "WHEN NOT MATCHED THEN "
            f"INSERT ({', '.join(insert_cols)}) "
            f"VALUES ({', '.join(f'source.{col}' for col in insert_cols)}) "
        )

    # End statement with ';'
    merge_sql_statement = merge_sql_statement.strip() + ";"

    return merge_sql_statement



from deprecated import deprecated

from spetlr.configurator.configurator import Configurator


@deprecated(
    reason="Don't import from spetlr.config_master. "
    "Use: from spetlr import Configurator.",
)
class TableConfigurator(Configurator):
    pass



from pyspark.sql import DataFrame

from .types import EtlBase, dataset_group


class Transformer(EtlBase):
    """If you only want to transform a single input dataframe,
    implement `process`
    if you want to transform a set of dataframes,
    implement `process_many`

    In regards to the etl step, the transformer CONSUMES the inputs
    and ADDs the result of its transformation stage.
    """

    def __init__(self, dataset_key: str = None):
        super().__init__()
        if dataset_key is None:
            dataset_key = type(self).__name__
        self.dataset_key = dataset_key

    def etl(self, inputs: dataset_group) -> dataset_group:
        if len(inputs) == 1:
            df = next(iter(inputs.values()))
            return {self.dataset_key: self.process(df)}
        return {self.dataset_key: self.process_many(inputs)}

    def process(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError()

    def process_many(self, datasets: dataset_group) -> DataFrame:
        raise NotImplementedError()



import sys
import time
import traceback
from datetime import datetime
from textwrap import dedent

import requests

from spetlr.exceptions import NoRunId

from .JobReflection import JobReflection


class SlackNotifier:
    """Send exceptions and notifications to slack webhooks
    while including stacktraces and links back to the job results.
    """

    def __init__(self, *webhookurl: str):
        self.webhookurls = list(webhookurl)

    def add_webhook_url(self, url: str):
        self.webhookurls.append(url)

    def notify_info(self, message: str = None, _stack_skip=1):
        """Send a notification to a webhook with the following example contents:
        "
        *A message was sent from your job <YOUR JOB NAME HERE>*

        Sent at 2022-11-20 01:03:17

        <YOUR MESSAGE HERE (IF ANY)>

        Caller info:
        ```
          File "/databricks/python_shell/scripts/PythonShell.py", line 29, in <module>
            launch_process()
        ...
        <STACK TRACE>
        ...
          File "/reporting/test_slack_reporting.py", line 53, in test_01_info_webhook
            slack.notify_info()
        ```
        <https://adb.../run/496|Go To Job Results>
        "
        """
        called_from = "".join(traceback.format_stack()[:-_stack_skip])
        try:
            job_name = JobReflection.get_job_name()
            text = f"*A message was sent from your job {job_name}*\n"
        except NoRunId:
            text = "*A message was sent from databricks*\n"

        text += f"\nSent at {self._slack_now()}\n"

        if message:
            text += f"\n{message}\n"

        text += f"\nCaller info:\n```\n{called_from}\n```\n"

        self._add_link_and_publish(text)

    def notify_exc(self):
        """Send a message about an ongoing exception
        to a webhook with the following example contents:
        "
        *An exception has occurred in your job <YOUR JOB NAME HERE>*

        The error occurred at 2022-11-20 01:03:17

        Traceback:
        ```
        Traceback (most recent call last):
          File "/reporting/test_slack_reporting.py", line 68, in test_02_exc_webhook
            1 / 0
        ZeroDivisionError: division by zero

        ```
        <https://adb-.../run/496|Go To Job Results>
        "
        """
        if sys.exc_info() == (None, None, None):
            return self.notify_info(
                message="SlackNotifier.notify_exc()"
                " was called with no active exception.",
                _stack_skip=2,
            )

        try:
            job_name = JobReflection.get_job_name()
            text = f"*An exception has occurred in your job {job_name}*\n"
        except NoRunId:
            text = "*An exception has occurred in databricks*\n"

        text += dedent(
            f"""The error occurred at {self._slack_now()}

        Traceback:
        ```
        {traceback.format_exc()}
        ```
        """
        )

        self._add_link_and_publish(text)

    def _slack_now(self):
        timestamp = int(time.time())
        date_format = "{date_long_pretty} {time_secs}"
        alt_text = str(datetime.now())
        return f"<!date^{timestamp}^{date_format}|{alt_text}>"

    def _add_link_and_publish(self, text: str):
        try:
            text += f"\n<{JobReflection.get_job_results_url()}|Go To Job Results>\n"
        except NoRunId:
            pass

        # send the whole thing off to the webhooks
        for url in self.webhookurls:
            res = requests.post(url=url, json={"text": text})
            if res.status_code != 200:
                # not raising an exception here makes this function more safe to call
                print(
                    "WARNING: Slack Notifier received "
                    f"unexpected status code {res.status_code}"
                )



"""
This function searches through .sql files and looks for statements to create tables
or databases:
https://spark.apache.org/docs/3.0.0-preview/sql-ref-syntax-ddl-create-database.html
https://spark.apache.org/docs/latest/sql-ref-syntax-ddl-create-table-datasource.html
and returns a dictionary of configuration details.
"""
import importlib.resources
from pathlib import Path
from types import ModuleType
from typing import Dict, Union

from spetlr.exceptions.configurator_exceptions import (
    SpetlrConfiguratorInvalidSqlCommentsException,
)

from .comments import _extract_comment_attributes
from .create import _walk_create_statement
from .init_sqlparse import parse


def _parse_sql_to_config(resource_path: Union[str, ModuleType]) -> Dict:
    details = {}
    for file_name in importlib.resources.contents(resource_path):
        extension = Path(file_name).suffix
        if not extension == ".sql":
            continue
        with importlib.resources.path(resource_path, file_name) as file_path:
            with open(file_path) as file:
                sql_code = file.read()

                # the sequence "-- COMMAND ----------" is used in jupyter notebooks
                # and separates cells.
                # We treat it as another way to end a statement
                sql_code = sql_code.replace("-- COMMAND ----------", ";")

                for statement in parse(sql_code):
                    comment_attributes = _extract_comment_attributes(statement)
                    if "key" not in comment_attributes:
                        # if no magic comments were used on the statement,
                        # then there is nothing to configure here
                        continue
                    table_id = comment_attributes.pop("key")

                    object_details = _walk_create_statement(statement)

                    for key in object_details:
                        if key in comment_attributes:
                            raise SpetlrConfiguratorInvalidSqlCommentsException(
                                f"Error for {key} in {table_id}, "
                                "The comments must not specify attributes, "
                                "that can be derived directly from the sql."
                            )
                    object_details.update(comment_attributes)
                    details[table_id] = object_details
    return details



from .extractor import Extractor
from .loader import Loader
from .orchestrator import Orchestrator
from .transformer import Transformer
from .transformer_nc import TransformerNC
from .types import EtlBase, dataset_group

__all__ = [
    "Loader",
    "Extractor",
    "Transformer",
    "Orchestrator",
    "EtlBase",
    "TransformerNC",
    "dataset_group",
]



import pyspark.sql.functions as f
from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.sql.SqlServer import SqlServer
from spetlr.utils import SelectAndCastColumns


class SimpleSqlServerTransformer(Transformer):
    def __init__(self, *, table_id: str, server: SqlServer, ignoreCase=False):
        super().__init__()
        self.server = server
        self.table_id = table_id
        self.ignoreCase = ignoreCase

    def process(self, df: DataFrame) -> DataFrame:
        # This transformation ensures that the selected columns
        # are casted to the target types
        # [f.col("ColumnName).cast("string").alias("ColumnName"), ...]

        df = SelectAndCastColumns(
            df=df,
            schema=self.server.read_table(self.table_id).schema,
            caseInsensitiveMatching=self.ignoreCase,
        )

        # If the format is timestamp, the seconds should be trunc
        col_choose = [
            f.date_trunc("second", f.col(x[0])).alias(x[0])
            if x[1] == "timestamp"
            else f.col(x[0])
            for x in df.dtypes
        ]
        df = df.select(col_choose)

        return df



"""Simplified access to the databricks api."""

import json
import sys

from databricks_api import DatabricksAPI

from spetlr.functions import init_dbutils


def getDbApi() -> DatabricksAPI:
    """
    This method automatically configures a databricks API client.
    In local running, the databricks-cli is used for configuration.
    Running on a cluster, the configuration is extracted from the job context.
    """
    try:
        dbutils = init_dbutils()
    except ModuleNotFoundError:
        # probably we are not in notebook
        dbutils = None

    if dbutils:
        context = json.loads(
            dbutils.notebook.entry_point.getDbutils().notebook().getContext().toJson()
        )
        host = context["extraContext"]["api_url"]
        token = context["extraContext"]["api_token"]
    else:
        try:
            from databricks_cli.configure.provider import ProfileConfigProvider
        except ModuleNotFoundError:
            print(
                "In local running, databricks-cli needs to be installed.",
                file=sys.stderr,
            )
            raise
        cfg = ProfileConfigProvider().get_config()
        host = cfg.host
        token = cfg.token

    if not host or not token:
        raise Exception("Unable to auto-configure api client.")

    return DatabricksAPI(host=host, token=token)



from typing import List, Union

from pyspark.sql import DataFrame

from spetlr.etl import TransformerNC


class DropColumnsTransformerNC(TransformerNC):
    def __init__(
        self,
        *,
        columnList: List[str],
        dataset_input_keys: Union[str, List[str]] = None,
        dataset_output_key: str = None,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
        )
        self.columnList = columnList

    def process(self, df: DataFrame) -> DataFrame:
        return df.drop(*self.columnList)



from typing import List

from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.utils import DropOldestDuplicates


class DropOldestDuplicatesTransformer(Transformer):
    def __init__(self, *, cols: List[str], orderByColumn: str):
        super().__init__()
        self.cols = cols
        self.orderByColumn = orderByColumn

    def process(self, df: DataFrame) -> DataFrame:
        return DropOldestDuplicates(
            df=df, cols=self.cols, orderByColumn=self.orderByColumn
        )



from decimal import Decimal

from pyspark.sql import DataFrame
from pyspark.sql.types import (
    DecimalType,
    DoubleType,
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase
from spetlrtools.time import dt_utc

from spetlr import Configurator
from spetlr.etl.loaders import SimpleLoader
from spetlr.functions import get_unique_tempview_name
from spetlr.spark import Spark
from spetlr.transformers.simple_sql_transformer import SimpleSqlServerTransformer
from spetlr.utils import DataframeCreator
from tests.cluster.sql.DeliverySqlServer import DeliverySqlServer


class SimpleSqlServerETLTests(DataframeTestCase):
    tc = None
    sql_server = None
    table_name = "dbo.Test1" + get_unique_tempview_name()
    table_id = "TableId1"

    @classmethod
    def setUpClass(cls):
        cls.sql_server = DeliverySqlServer()
        cls.tc = Configurator()
        cls.tc.clear_all_configurations()

        # Register the delivery table for the table configurator
        cls.tc.register(
            cls.table_id,
            {
                "name": cls.table_name,
            },
        )
        cls.tc.reset(debug=True)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.sql_server.drop_table(cls.table_id)
        cls.tc.reset(debug=False)

    def test01_can_transform(self):
        self.create_test_table()
        df = self.create_data()

        # Test that timestamp before has miliseconds also
        date_test = df.select("testcolumn4").collect()[0][0]
        self.assertEqual(
            dt_utc(2021, 1, 1, 14, 45, 22, 32).replace(tzinfo=None), date_test
        )

        df_out = SimpleSqlServerTransformer(
            table_id=self.table_id, server=self.sql_server
        ).process(df)

        # Check that timestamp are truncated down to only seconds
        # The 32 miliseconds should be gone.
        date_test = df_out.select("testcolumn4").collect()[0][0]
        self.assertEqual(dt_utc(2021, 1, 1, 14, 45, 22).replace(tzinfo=None), date_test)

        schema_expected = StructType(
            [
                StructField("testcolumn", IntegerType(), True),
                StructField("testcolumn2", DecimalType(12, 3), True),
                StructField("testcolumn3", StringType(), True),
                StructField("testcolumn4", TimestampType(), True),
            ]
        )

        self.assertEqual(df_out.schema, schema_expected)

    def test02_can_transform_and_load(self):
        # mix up the column spelling to test case-insensitive matching this time
        df = Spark.get().createDataFrame(
            [(123, 1001.322, "Hello", dt_utc(2021, 1, 1, 14, 45, 22, 32))],
            """
                TeStCoLuMn INT,
                TESTcolumn2 DOUBLE,
                testCOLUMN3 STRING,
                TestColumn4 TIMESTAMP
            """,
        )

        # Use transformer
        df_out = SimpleSqlServerTransformer(
            table_id=self.table_id, server=self.sql_server, ignoreCase=True
        ).process(df)

        SimpleLoader(handle=self.sql_server.from_tc(self.table_id)).save(df_out)

        df_with_data = self.sql_server.read_table(self.table_id)

        # Test that the datarow can be read
        self.assertDataframeMatches(
            df_with_data,
            None,
            [(123, Decimal("1001.322"), "Hello", dt_utc(2021, 1, 1, 14, 45, 22))],
        )

    def create_test_table(self):
        sql_argument = f"""
                IF OBJECT_ID('{self.table_name}', 'U') IS NULL
                BEGIN
                CREATE TABLE {self.table_name}
                (
                testcolumn INT NULL,
                testcolumn2 DECIMAL(12,3) NULL,
                testcolumn3 [nvarchar](50) NULL,
                testcolumn4 DATETIME NULL
                )
                END
                                """
        self.sql_server.execute_sql(sql_argument)

    def create_data(self) -> DataFrame:
        schema = StructType(
            [
                StructField("testcolumn", IntegerType(), True),
                StructField("testcolumn2", DoubleType(), True),
                StructField("testcolumn3", StringType(), True),
                StructField("testcolumn4", TimestampType(), True),
            ]
        )
        cols = ["testcolumn", "testcolumn2", "testcolumn3", "testcolumn4"]

        insert_data = (123, 1001.322, "Hello", dt_utc(2021, 1, 1, 14, 45, 22, 32))

        df_new = DataframeCreator.make_partial(
            schema=schema, columns=cols, data=[insert_data]
        )

        return df_new.orderBy("testcolumn")



import unittest
from unittest.mock import Mock

from spetlrtools.time import dt_utc

from spetlr import Configurator
from spetlr.delta import DeltaHandle
from spetlr.eh import EventHubCaptureExtractor
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaExtractor import (
    EhJsonToDeltaExtractor,
)
from spetlr.spark import Spark


class JsonEhOrchestratorUnitTests(unittest.TestCase):
    tc: Configurator

    @classmethod
    def setUpClass(cls) -> None:
        cls.tc = Configurator()
        cls.tc.set_debug()
        cls.tc.clear_all_configurations()
        cls.tc.register("TblYMD", {"name": "TableYMD{ID}"})
        cls.tc.register("TblYMDH", {"name": "TableYMDH{ID}"})
        cls.tc.register("TblPdate", {"name": "TablePdate{ID}"})

        spark = Spark.get()
        spark.sql(f"DROP TABLE IF EXISTS {cls.tc.table_name('TblYMD')}")
        spark.sql(f"DROP TABLE IF EXISTS {cls.tc.table_name('TblYMDH')}")
        spark.sql(f"DROP TABLE IF EXISTS {cls.tc.table_name('TblPdate')}")

        spark.sql(
            f"""
            CREATE TABLE {cls.tc.table_name('TblYMD')}
            (id int, name string, y int, m int, d int)
            PARTITIONED BY (y,m,d)
        """
        )
        spark.sql(
            f"""
            CREATE TABLE {cls.tc.table_name('TblYMDH')}
            (id int, name string, y int, m int, d int, h int)
            PARTITIONED BY (y,m,d,h)
        """
        )
        spark.sql(
            f"""
            CREATE TABLE {cls.tc.table_name('TblPdate')}
            (id int, name string, pdate timestamp)
            PARTITIONED BY (pdate)
        """
        )

    def test_reading_YMD(self):
        dh = DeltaHandle.from_tc("TblYMD")
        dh.truncate()
        dh.append(
            Spark.get().createDataFrame(
                [
                    (42, "spam", 2019, 6, 23),
                    # this is where the eventhub should read from:
                    (84, "eggs", 2020, 9, 5),
                ],
                dh.read().schema,
            )
        )
        eh = Mock(EventHubCaptureExtractor)
        eh.get_partitioning.side_effect = lambda: ["y", "m", "d"]

        EhJsonToDeltaExtractor(dh=dh, eh=eh).read()

        # check that the truncation worked
        self.assertEqual(len(dh.read().collect()), 1)

        # check that the from_partition argument is as expected
        eh.read.assert_called_once_with(from_partition=dt_utc(2020, 9, 5))

    def test_reading_YMDH(self):
        dh = DeltaHandle.from_tc("TblYMDH")
        dh.truncate()
        dh.append(
            Spark.get().createDataFrame(
                [
                    (42, "spam", 2019, 6, 23, 15),
                    # this is where the eventhub should read from:
                    (84, "eggs", 2020, 9, 5, 12),
                ],
                dh.read().schema,
            )
        )
        eh = Mock(EventHubCaptureExtractor)
        eh.get_partitioning.side_effect = lambda: ["y", "m", "d", "h"]

        EhJsonToDeltaExtractor(dh=dh, eh=eh).read()

        # check that the truncation worked
        self.assertEqual(len(dh.read().collect()), 1)

        # check that the from_partition argument is as expected
        eh.read.assert_called_once_with(from_partition=dt_utc(2020, 9, 5, 12))

    def test_reading_Pdate(self):
        dh = DeltaHandle.from_tc("TblPdate")
        dh.truncate()
        dh.append(
            Spark.get().createDataFrame(
                [
                    (42, "spam", dt_utc(2019, 6, 23, 15)),
                    # this is where the eventhub should read from:
                    (84, "eggs", dt_utc(2020, 9, 5, 12)),
                ],
                dh.read().schema,
            )
        )
        eh = Mock(EventHubCaptureExtractor)
        eh.get_partitioning.side_effect = lambda: ["y", "m", "d", "h"]

        EhJsonToDeltaExtractor(dh=dh, eh=eh).read()

        # check that the truncation worked
        self.assertEqual(len(dh.read().collect()), 1)

        # check that the from_partition argument is as expected
        eh.read.assert_called_once_with(from_partition=dt_utc(2020, 9, 5, 12))



from datetime import date, datetime
from typing import Any, List, Union

from pyspark.sql import DataFrame

from spetlr.exceptions import SpetlrException
from spetlr.sql.SqlBaseServer import SqlBaseServer
from spetlr.tables.TableHandle import TableHandle


class SqlHandleException(SpetlrException):
    pass


class SqlHandleInvalidName(SqlHandleException):
    pass


class SqlHandleInvalidFormat(SqlHandleException):
    pass


class SqlHandle(TableHandle):
    def __init__(self, name: str, sql_server: SqlBaseServer):
        self._name = name
        self._sql_server = sql_server

        self._validate()

    def _validate(self):
        """Validates that the name is either db.table or just table."""
        name_parts = self._name.split(".")
        if len(name_parts) == 1:
            self._db = None
            self._table_name = name_parts[0]
        elif len(name_parts) == 2:
            self._db = name_parts[0]
            self._table_name = name_parts[1]
        else:
            raise SqlHandleInvalidName(f"Could not parse name {self._name}")

    def read(self) -> DataFrame:
        """Read table by path if location is given, otherwise from name."""
        return self._sql_server.read_table_by_name(table_name=self._name)

    def write_or_append(self, df: DataFrame, mode: str) -> None:
        assert mode in {"append", "overwrite"}

        append = True if mode == "append" else False

        return self._sql_server.write_table_by_name(
            df_source=df, table_name=self._name, append=append
        )

    def overwrite(self, df: DataFrame) -> None:
        return self._sql_server.write_table_by_name(
            df_source=df, table_name=self._name, append=False
        )

    def append(self, df: DataFrame) -> None:
        return self._sql_server.write_table_by_name(
            df_source=df, table_name=self._name, append=True
        )

    def upsert(self, df: DataFrame, join_cols: List[str]) -> Union[DataFrame, None]:
        return self._sql_server.upsert_to_table_by_name(
            df_source=df, table_name=self._name, join_cols=join_cols
        )

    def truncate(self) -> None:
        self._sql_server.truncate_table_by_name(table_name=self._name)

    def drop(self) -> None:
        self._sql_server.drop_table_by_name(table_name=self._name)

    def drop_and_delete(self) -> None:
        self.drop()

    def get_tablename(self) -> str:
        return self._name

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        needs_quotes = (
            isinstance(comparison_limit, str)
            or isinstance(comparison_limit, date)
            or isinstance(comparison_limit, datetime)
        )
        limit = f"'{comparison_limit}'" if needs_quotes else comparison_limit
        sql_str = (
            f"DELETE FROM {self._name}"
            f" WHERE {comparison_col} {comparison_operator} {limit};"
        )
        self._sql_server.execute_sql(sql_str)



import unittest

from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.etl import TransformerNC
from spetlr.etl.types import dataset_group
from spetlr.spark import Spark


class TransformerNCTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.transformer = TestTransformer()
        cls.df = create_dataframe()

        cls.transformer_output_key = TestTransformer(dataset_output_key="my_output_key")
        cls.transformer_input_key = TestTransformer(dataset_input_keys="my_input_key")
        cls.transformer_input_key_list = TestTransformer(
            dataset_input_keys=["my_input_key_1", "my_input_key_2"]
        )

    def test_process(self):
        # process is called when:
        # - inputs has one df
        # - dataset_input_key is not given
        result = self.transformer.etl({"df": self.df})
        self.assertIs(self.df, list(result.values())[1])

    def test_process_many(self):
        # process_many is called when:
        # - inputs has more than one df
        # - dataset_input_key is not given
        result = self.transformer.etl({"df1": self.df, "df2": self.df})
        self.assertEqual(6, list(result.values())[2].count())

    def test_process_no_key(self):
        # process sets dataset_ouput_key equivalent to the transformer class when:
        # - dataset_output_key is not given
        result = self.transformer.etl({"df": self.df})
        self.assertEqual("TestTransformer", list(result.keys())[1])

    def test_process_many_no_key(self):
        # process_many sets dataset_ouput_key equivalent to the transformer class when:
        # - dataset_output_key is not given
        result = self.transformer.etl({"df1": self.df, "df2": self.df})
        self.assertEqual("TestTransformer", list(result.keys())[2])

    def test_process_output_key(self):
        # process sets dataset_output_key equiavalent to the input when:
        # - dataset_output_key is given
        result = self.transformer_output_key.etl({"df": self.df})
        self.assertEqual("my_output_key", list(result.keys())[1])

    def test_process_many_output_key(self):
        # process_many sets dataset_output_key equiavalent to the input when:
        # - dataset_output_key is given
        result = self.transformer_output_key.etl({"df1": self.df, "df2": self.df})
        self.assertEqual("my_output_key", list(result.keys())[2])

    def test_process_input_key(self):
        # process is called when:
        # - dataset_input_key is given
        result = self.transformer_input_key.etl({"my_input_key": self.df})
        self.assertIs(self.df, list(result.values())[1])

        with self.assertRaises(KeyError):
            # error when:
            # - the provided dataset_input_key does not exist in inputs
            self.transformer_input_key.etl({"df": self.df})

    def test_process_many_input_key_list(self):
        # process_many is called when:
        # - dataset_input_key_list is given
        result = self.transformer_input_key_list.etl(
            {"my_input_key_1": self.df, "my_input_key_2": self.df}
        )
        self.assertEqual(6, list(result.values())[2].count())

        with self.assertRaises(AssertionError):
            # error when:
            # - the provided dataset_input_key_list contains an incorrect key
            self.transformer_input_key_list.etl({"df1": self.df, "df2": self.df})


class TestTransformer(TransformerNC):
    def process(self, df: DataFrame) -> DataFrame:
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        assert len(datasets) == 2
        df1, df2 = list(datasets.values())
        return df1.union(df2)


def create_dataframe():
    data = [(1, "1"), (2, "2"), (3, "3")]

    schema = StructType(
        [
            StructField("id", IntegerType(), False),
            StructField("text", StringType(), False),
        ]
    )

    return Spark.get().createDataFrame(data=data, schema=schema)


if __name__ == "__main__":
    unittest.main()



from .EventHubCapture import EventHubCapture  # noqa: F401
from .EventHubCaptureExtractor import EventHubCaptureExtractor  # noqa: F401
from .EventHubJsonPublisher import EventHubJsonPublisher  # noqa: F401
from .EventHubStream import EventHubStream  # noqa: F401



from typing import List, Union

from pyspark.sql import DataFrame

from spetlr.etl import TransformerNC
from spetlr.etl.types import dataset_group
from spetlr.exceptions import (
    ColumnDoesNotExistException,
    MoreThanTwoDataFramesException,
)


class JoinDataframesTransformerNC(TransformerNC):
    """
    This transformer joins two DataFrames together.

    Attributes:
    ----------
        first_dataframe_join_key : str
            the name of the column that will be joined on in the first DataFrame
        second_dataframe_join_key : str
            the name of the column that will be joined on in the second DataFrame
        join_type : str
            the type of the join
        dataset_input_keys : Union[str, List[str]]
            list of dataset keys
        dataset_output_key : str
            identifier for the output DataFrame

    Methods
    -------
    process_many(dataset: dataset_group):
        returns the two input DataFrames joined together
    """

    def __init__(
        self,
        first_dataframe_join_key: str,
        second_dataframe_join_key: str,
        join_type: str = "inner",
        dataset_input_keys: Union[str, List[str]] = None,
        dataset_output_key: str = None,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys, dataset_output_key=dataset_output_key
        )
        self.first_dataframe_join_key = first_dataframe_join_key
        self.second_dataframe_join_key = second_dataframe_join_key
        self.join_type = join_type

    def process_many(self, dataset: dataset_group) -> DataFrame:
        """
        Takes a dictionary of datasets and joins two specified datasets together.
        """

        if len(self.dataset_input_key_list) > 2:
            raise MoreThanTwoDataFramesException(
                """More than two DataFrames are specified in 'dataset_input_keys'.
                This transformer can only join two DataFrames at a time."""
            )

        first_df = dataset[self.dataset_input_key_list[0]]
        second_df = dataset[self.dataset_input_key_list[1]]

        if self.first_dataframe_join_key not in first_df.columns:
            raise ColumnDoesNotExistException(
                f"""'{self.first_dataframe_join_key}' cannot be found in
                the first DataFrame."""
            )

        if self.second_dataframe_join_key not in second_df.columns:
            raise ColumnDoesNotExistException(
                f"""'{self.second_dataframe_join_key}' cannot be found in
                the second DataFrame."""
            )

        return first_df.join(
            second_df,
            first_df[self.first_dataframe_join_key]
            == second_df[self.second_dataframe_join_key],
            how=self.join_type,
        )



#!/usr/bin/env python
#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""Entrypoint module for `python -m sqlparse`.

Why does this file exist, and why __main__? For more info, read:
- https://www.python.org/dev/peps/pep-0338/
- https://docs.python.org/2/using/cmdline.html#cmdoption-m
- https://docs.python.org/3/using/cmdline.html#cmdoption-m
"""

import sys

from spetlr.configurator.sql.sqlparse.cli import main

if __name__ == "__main__":
    sys.exit(main())



from datetime import date, datetime
from typing import Any, List, Optional, Union

from pyspark.sql import DataFrame

from spetlr.configurator.configurator import Configurator
from spetlr.exceptions import SpetlrException
from spetlr.functions import get_unique_tempview_name, init_dbutils
from spetlr.spark import Spark
from spetlr.tables.TableHandle import TableHandle
from spetlr.utils.CheckDfMerge import CheckDfMerge
from spetlr.utils.GetMergeStatement import GetMergeStatement


class DeltaHandleException(SpetlrException):
    pass


class DeltaHandleInvalidName(DeltaHandleException):
    pass


class DeltaHandleInvalidFormat(DeltaHandleException):
    pass


class DeltaHandle(TableHandle):
    def __init__(self, name: str, location: str = None, data_format: str = "delta"):
        self._name = name
        self._location = location
        self._data_format = data_format

        self._partitioning: Optional[List[str]] = None

        self._validate()

    @classmethod
    def from_tc(cls, id: str) -> "DeltaHandle":
        tc = Configurator()
        return cls(
            name=tc.table_property(id, "name", ""),
            location=tc.table_property(id, "path", ""),
            data_format=tc.table_property(id, "format", "delta"),
        )

    def _validate(self):
        """Validates that the name is either db.table or just table."""
        if not self._name:
            if not self._location:
                raise DeltaHandleInvalidName(
                    "Cannot create DeltaHandle without name or path"
                )
            self._name = f"delta.`{self._location}`"
        else:
            name_parts = self._name.split(".")
            if len(name_parts) == 1:
                self._db = None
                self._table_name = name_parts[0]
            elif len(name_parts) == 2:
                self._db = name_parts[0]
                self._table_name = name_parts[1]
            else:
                raise DeltaHandleInvalidName(f"Could not parse name {self._name}")

        # only format delta is supported.
        if self._data_format != "delta":
            raise DeltaHandleInvalidFormat("Only format delta is supported.")

    def read(self) -> DataFrame:
        """Read table by path if location is given, otherwise from name."""
        if self._location:
            return Spark.get().read.format(self._data_format).load(self._location)
        return Spark.get().table(self._name)

    def write_or_append(
        self, df: DataFrame, mode: str, mergeSchema: bool = None
    ) -> None:
        assert mode in {"append", "overwrite"}

        writer = df.write.format(self._data_format).mode(mode)
        if mergeSchema is not None:
            writer = writer.option("mergeSchema", "true" if mergeSchema else "false")

        if self._location:
            return writer.save(self._location)

        return writer.saveAsTable(self._name)

    def overwrite(self, df: DataFrame, mergeSchema: bool = None) -> None:
        return self.write_or_append(df, "overwrite", mergeSchema=mergeSchema)

    def append(self, df: DataFrame, mergeSchema: bool = None) -> None:
        return self.write_or_append(df, "append", mergeSchema=mergeSchema)

    def truncate(self) -> None:
        Spark.get().sql(f"TRUNCATE TABLE {self._name};")

    def drop(self) -> None:
        Spark.get().sql(f"DROP TABLE IF EXISTS {self._name};")

    def drop_and_delete(self) -> None:
        self.drop()
        if self._location:
            init_dbutils().fs.rm(self._location, True)

    def create_hive_table(self) -> None:
        sql = f"CREATE TABLE IF NOT EXISTS {self._name} "
        if self._location:
            sql += f" USING DELTA LOCATION '{self._location}'"
        Spark.get().sql(sql)

    def recreate_hive_table(self):
        self.drop()
        self.create_hive_table()

    def get_partitioning(self):
        """The result of DESCRIBE TABLE tablename is like this:
        +-----------------+---------------+-------+
        |         col_name|      data_type|comment|
        +-----------------+---------------+-------+
        |           mycolA|         string|       |
        |           myColB|            int|       |
        |                 |               |       |
        |   # Partitioning|               |       |
        |           Part 0|         mycolA|       |
        +-----------------+---------------+-------+
        but this method return the partitioning in the form ['mycolA'],
        if there is no partitioning, an empty list is returned.
        """
        if self._partitioning is None:
            # create an iterator object and use it in two steps
            rows_iter = iter(
                Spark.get().sql(f"DESCRIBE TABLE {self.get_tablename()}").collect()
            )

            # roll over the iterator until you see the title line
            for row in rows_iter:
                # discard rows until the important section header
                if row.col_name.strip() == "# Partitioning":
                    break
            # at this point, the iterator has moved past the section heading
            # leaving only the rows with "Part 1" etc.

            # create a list from the rest of the iterator like [(0,colA), (1,colB)]
            parts = [
                (int(row.col_name[5:]), row.data_type)
                for row in rows_iter
                if row.col_name.startswith("Part ")
            ]
            # sort, just in case the parts were out of order.
            parts.sort()

            # discard the index and put into an ordered list.
            self._partitioning = [p[1] for p in parts]
        return self._partitioning

    def get_tablename(self) -> str:
        return self._name

    def upsert(
        self,
        df: DataFrame,
        join_cols: List[str],
    ) -> Union[DataFrame, None]:
        if df is None:
            return None

        df = df.filter(" AND ".join(f"({col} is NOT NULL)" for col in join_cols))
        print(
            "Rows with NULL join keys found in input dataframe"
            " will be discarded before load."
        )

        df_target = self.read()

        # If the target is empty, always do faster full load
        if len(df_target.take(1)) == 0:
            return self.write_or_append(df, mode="overwrite")

        # Find records that need to be updated in the target (happens seldom)

        # Define the column to be used for checking for new rows
        # Checking the null-ness of one right row is sufficient to mark the row as new,
        # since null keys are disallowed.

        df, merge_required = CheckDfMerge(
            df=df,
            df_target=df_target,
            join_cols=join_cols,
            avoid_cols=[],
        )

        if not merge_required:
            return self.write_or_append(df, mode="append")

        temp_view_name = get_unique_tempview_name()
        df.createOrReplaceTempView(temp_view_name)

        target_table_name = self.get_tablename()
        non_join_cols = [col for col in df.columns if col not in join_cols]

        merge_sql_statement = GetMergeStatement(
            merge_statement_type="delta",
            target_table_name=target_table_name,
            source_table_name=temp_view_name,
            join_cols=join_cols,
            insert_cols=df.columns,
            update_cols=non_join_cols,
            special_update_set="",
        )

        Spark.get().sql(merge_sql_statement)

        print("Incremental Base - incremental load with merge")

        return df

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        needs_quotes = (
            isinstance(comparison_limit, str)
            or isinstance(comparison_limit, date)
            or isinstance(comparison_limit, datetime)
        )
        limit = f"'{comparison_limit}'" if needs_quotes else comparison_limit
        sql_str = (
            f"DELETE FROM {self._name}"
            f" WHERE {comparison_col} {comparison_operator} {limit};"
        )
        Spark.get().sql(sql_str)
